package ru.dbotthepony.mc.otm.container

import it.unimi.dsi.fastutil.ints.IntCollection
import net.minecraft.world.item.ItemStack
import ru.dbotthepony.kommons.collect.iterateClearBits
import ru.dbotthepony.kommons.collect.iterateSetBits
import ru.dbotthepony.mc.otm.core.collect.IntRange2Set
import ru.dbotthepony.mc.otm.core.collect.map
import java.util.*

abstract class BitmapTrackingContainer<out S : IContainerSlot> : IEnhancedContainer<S> {
	protected val bitmap = BitSet()

	final override fun isEmpty(): Boolean {
		return bitmap.isEmpty
	}

	final override fun nextEmptySlot(startIndex: Int): Int {
		if (startIndex >= containerSize)
			return -1
		else if (startIndex < 0)
			return bitmap.nextClearBit(0)
		else
			return bitmap.nextClearBit(startIndex)
	}

	final override fun nextNonEmptySlot(startIndex: Int): Int {
		if (startIndex >= containerSize)
			return -1
		else if (startIndex < 0)
			return bitmap.nextSetBit(0)
		else
			return bitmap.nextSetBit(startIndex)
	}

	final override fun iterator(): Iterator<ItemStack> {
		return bitmap.iterateSetBits(containerSize).map { this[it] }
	}

	final override fun nonEmptySlotIndexIterator(): IntIterator {
		return bitmap.iterateSetBits(containerSize)
	}

	final override fun emptySlotIndexIterator(): IntIterator {
		return bitmap.iterateClearBits(containerSize)
	}

	final override fun emptySlotIndexIterator(allowedSlots: IntCollection): IntIterator {
		if (allowedSlots is IntRange2Set && allowedSlots.isNotEmpty())
			return bitmap.iterateClearBits(allowedSlots.firstInt(), allowedSlots.lastInt() + 1)

		return super.emptySlotIndexIterator(allowedSlots)
	}

	final override fun nonEmptySlotIndexIterator(allowedSlots: IntCollection): IntIterator {
		if (allowedSlots is IntRange2Set && allowedSlots.isNotEmpty())
			return bitmap.iterateSetBits(allowedSlots.firstInt(), allowedSlots.lastInt() + 1)

		return super.nonEmptySlotIndexIterator(allowedSlots)
	}
}
